"""This module defines functions for solving linear assignment problems."""

from numpy import arange, insert, delete, zeros, unique, vstack, ceil, argmin, tile
from scipy.optimize import linear_sum_assignment as lap

__all__ = ['multilap', 'SolutionDepletionException']

class SolutionDepletionException(Exception):
    pass

def expand_nodes(node):
    includes0, excludes0, solution = node

    pairs = []
    R, C = solution
    for r, c in zip(R, C):
        pairs.append((r, c))

    nodes = []; previous_pairs = []
    for i, pair in enumerate(pairs):
        if i == len(pairs) - 1:
            continue
        if pair in includes0:
            continue
        if pair in excludes0: # unlikely
            raise ValueError('%s should be excluded'%str(pair))
        
        includes = list(includes0)
        includes.extend(previous_pairs)

        excludes = list(excludes0)
        excludes.append(pair)

        nodes.append([includes, excludes, None])
        previous_pairs.append(pair)

    return nodes
    
def multilap(cost_matrix, nodes=[], BIG_NUMBER=1e6):
    """ Finds the (next) optimal solution to the linear assignment problem. 
    The function can handle the cases where each row can be assigned to multiple 
    columns. 

    .. [KM68] Murty KG. Letter to the editor-An algorithm for ranking 
       all the assignments in order of increasing cost.
       *Operations research* **1968** 16(3):682-687.
    """

    m, n = cost_matrix.shape
    row_labels = arange(m)
    
    # make duplicates for alternative assignments
    n_copies = int(ceil(float(n) / m))
    if n_copies > 1:
        cost_matrix = vstack((cost_matrix,)*n_copies)
        row_labels = tile(row_labels, n_copies)
    else:
        cost_matrix = cost_matrix.copy()

    if len(nodes):
        M, N = cost_matrix.shape

        costs = []
        for node in nodes:
            includes, excludes, assignments = node
            if assignments is None:
                D = cost_matrix.copy()
                R = arange(M)
                C = arange(N)

                # mask excludes
                for r, c in excludes:
                    r_ = row_labels[r]
                    D[row_labels==r_, c] = BIG_NUMBER
                
                # remove includes
                delR = []; delC = []
                for r, c in includes:
                    delR.append(r)
                    delC.append(c)

                if delC:
                    D = delete(D, delC, axis=1)
                    C = delete(C, delC)
                
                if delR:
                    D = delete(D, delR, axis=0)
                    R = delete(R, delR)
                
                # solve lap
                I, J = lap(D)
                cost = D[I, J].sum()
                
                if includes:
                    cost += cost_matrix[delR, delC].sum()
                row_ind = R[I]; col_ind = C[J]
                # add back includes
                if includes:
                    row_ind = insert(row_ind, 0, delR)
                    col_ind = insert(col_ind, 0, delC)

                assignments = node[-1] = (row_ind, col_ind)
            else:
                R, C = assignments
                cost = cost_matrix[R, C].sum()
            
            costs.append(cost)
        
        # remove assignments with violations. -0.01 is to avoid precision problem
        for i in reversed(range(len(costs))):
            if costs[i] - BIG_NUMBER >= -0.01:
                costs.pop(i)
                nodes.pop(i)

        if len(costs) == 0:
            raise SolutionDepletionException('solution depleted')

        i = argmin(costs)
        node = nodes.pop(i)
        R, C = node[-1]
    else:
        R, C = lap(cost_matrix)
        node = ([], [], (R, C))
        cost = cost_matrix[R, C].sum()

    new_nodes = expand_nodes(node)
    nodes.extend(new_nodes)

    R_ = row_labels[R]

    return (R_, C), gen_mappings((R_, C))

def gen_mappings(assignments):
    from itertools import product as iproduct

    I, J = assignments
    m = len(unique(I))
    M = I.max() + 1

    if m == len(I):
        mappings = [(I, J)]
    else:
        pool = [[] for _ in range(M)]
        for i, j in zip(I, J):
            pool[i].append((i, j))

        mappings = []
        for mapping in iproduct(*pool):
            r = zeros(len(mapping), dtype=int)
            c = zeros(len(mapping), dtype=int)
            for i, pair in enumerate(mapping):
                r[i] = pair[0]
                c[i] = pair[1]
            mappings.append((r, c))
        
    return mappings



"""This module defines miscellaneous utility functions."""
import re

from numpy import unique, linalg, diag, sqrt, dot, zeros, allclose, abs, bytes_
from numpy import diff, where, insert, nan, isnan, loadtxt, array, round, average, min, max
from numpy import sign, arange, asarray, ndarray, subtract, power, sum, isscalar, empty, triu, tril, median
from numpy import all
from collections import Counter
import numbers

from prody import PY3K
from .logger import LOGGER

from Bio.Data import IUPACData

from xml.etree.ElementTree import Element

__all__ = ['Everything', 'Cursor', 'ImageCursor', 'rangeString', 'alnum', 'importLA', 'dictElement',
           'intorfloat', 'startswith', 'showFigure', 'countBytes', 'sqrtm',
           'saxsWater', 'count', 'addEnds', 'copy', 'dictElementLoop', 'index',
           'getDataPath', 'openData', 'chr2', 'toChararray', 'interpY', 'cmp', 'pystr',
           'getValue', 'indentElement', 'isPDB', 'isURL', 'isListLike', 'isSymmetric', 'makeSymmetric',
           'getDistance', 'fastin', 'createStringIO', 'div0', 'wmean', 'bin2dec', 'wrapModes', 
           'fixArraySize', 'decToHybrid36', 'hybrid36ToDec', 'DTYPE', 'checkIdentifiers', 'split', 'mad',
           'importDec', 'impLoadModule']

DTYPE = array(['a']).dtype.char  # 'S' for PY2K and 'U' for PY3K
CURSORS = []

# Note that the chain id can be blank (space). Examples:
# 3TT1, 3tt1A, 3tt1:A, 3tt1_A, 3tt1-A, 3tt1 A
isPDB = re.compile('^[A-Za-z0-9]{4}[ -_:]{,1}[A-Za-z0-9 ]{,1}$').match

# django url validation regex
isURL = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE).match

class Everything(object):
    """A place for everything."""

    def __contains__(self, what):

        return True

class Cursor(object):
    def __init__(self, ax):
        import matplotlib.patheffects as PathEffects

        self.ax = ax
        self.lx = ax.axhline(color='k', linestyle='--', linewidth=0.)  # the horiz line
        self.ly = ax.axvline(color='k', linestyle='--', linewidth=0.)  # the vert line

        # text location in axes coords
        self.txt = ax.text(0., 1., '', color='k', verticalalignment='bottom')
        self.txt.set_path_effects([PathEffects.withStroke(linewidth=3, foreground='w')])
        
        # preserve the cursor reference
        global CURSORS
        CURSORS.append(self)

    def onClick(self, event):
        from matplotlib.pyplot import draw

        if event.inaxes != self.ax:
            return

        if event.button == 1:
            self.show(event)
        elif event.button == 3:
            self.clear(event)

        draw()

    def show(self, event):
        x, y = event.xdata, event.ydata
        # update the line positions
        self.lx.set_ydata(y)
        self.ly.set_xdata(x)

        self.lx.set_linewidth(.75)
        self.ly.set_linewidth(.75)

        self.txt.set_text(' x=%1.2f, y=%1.2f' % (x, y))
        self.txt.set_position((x, y))

    def clear(self, event):
        self.lx.set_linewidth(0.)
        self.ly.set_linewidth(0.)

        self.txt.set_text('')

class ImageCursor(Cursor):
    def __init__(self, ax, image, atoms=None):
        super(ImageCursor, self).__init__(ax)
        self.image = image
        self.atoms = atoms
    
    def show(self, event):
        x, y = event.xdata, event.ydata
        # update the line positions
        self.lx.set_ydata(y)
        self.ly.set_xdata(x)

        self.lx.set_linewidth(1.)
        self.ly.set_linewidth(1.)

        data = self.get_cursor_data(event)
        if data is None:
            return
            
        i, j, v = data

        if v > 1e-4 and v < 1e4:
            template = ' x=%s, y=%s [%s]'
        else:
            template = ' x=%s, y=%s [%s]'
        if self.atoms is None:
            self.txt.set_text(template % (j, i, v)) 
        else:
            if not isListLike(self.atoms):
                seq = self.atoms.getSequence()
                resnums = self.atoms.getResnums()

                a = seq[j] + str(resnums[j])
                b = seq[i] + str(resnums[i])
                
            else:
                xseq = self.atoms[0].getSequence()
                xresnums = self.atoms[0].getResnums()

                yseq = self.atoms[1].getSequence()
                yresnums = self.atoms[1].getResnums()

                a = xseq[j] + str(xresnums[j])
                b = yseq[i] + str(yresnums[i])

            self.txt.set_text(template % (a, b, str(v)))
        self.txt.set_position((x, y))

    def get_cursor_data(self, event):
        """Get the cursor data for a given event"""
        from matplotlib.transforms import Bbox, BboxTransform

        aximg = self.image
        xmin, xmax, ymin, ymax = aximg.get_extent()
        if aximg.origin == 'upper':
            ymin, ymax = ymax, ymin

        arr = aximg.get_array()
        data_extent = Bbox([[ymin, xmin], [ymax, xmax]])
        array_extent = Bbox([[0, 0], arr.shape[:2]])
        trans = BboxTransform(boxin=data_extent, boxout=array_extent)
        y, x = event.ydata, event.xdata
        point = trans.transform_point([y, x])
        if any(isnan(point)):
            return None
        i, j = point.astype(int)
        # Clip the coordinates at array bounds
        if not (0 <= i < arr.shape[0]) or not (0 <= j < arr.shape[1]):
            return None
        else:
            return i, j, arr[i, j]

def rangeString(lint, sep=' ', rng=' to ', exc=False, pos=True):
    """Returns a structured string for a given list of integers.

    :arg lint: integer list or array
    :arg sep: range or number separator
    :arg rng: range symbol
    :arg exc: set **True** if range symbol is exclusive
    :arg pos: only consider zero and positive integers

    .. ipython:: python

       from prody.utilities import rangeString
       lint = [1, 2, 3, 4, 10, 15, 16, 17]
       rangeString(lint)
       rangeString(lint, sep=',', rng='-')
       rangeString(lint, ',', ':', exc=True)"""

    ints = unique(lint)
    if len(ints) == 0:
        return ''
    if pos and ints[0] < 0:
        ints = ints[ints > -1]

    prev = ints[0]
    lint = [[prev]]
    for i in ints[1:]:
        if i - prev > 1:
            lint.append([i])
        else:
            lint[-1].append(i)
        prev = i
    exc = int(exc)
    return sep.join([str(l[0]) if len(l) == 1 else
                     str(l[0]) + rng + str(l[-1] + exc) for l in lint])


def alnum(string, alt='_', trim=False, single=False):
    """Replace non alpha numeric characters with *alt*.  If *trim* is **True**
    remove preceding and trailing *arg* characters.  If *single* is **True**,
    contain only a single joining *alt* character. """

    result = ''
    multi = not bool(single)
    prev = None
    for char in string:
        if char.isalnum():
            result += char
            prev = char
        else:
            if multi or prev != alt:
                result += alt
            prev = alt
    trim = int(bool(trim))
    result = result[trim * (result[0] == alt):
                    len(result) - trim * (result[-1] == alt)]
    return result


def importLA():
    """Returns one of :mod:`scipy.linalg` or :mod:`numpy.linalg`."""

    try:
        import scipy.linalg as linalg
    except ImportError:
        try:
            import numpy.linalg as linalg
        except:
            raise ImportError('scipy.linalg or numpy.linalg is required for '
                              'NMA and structure alignment calculations')
    return linalg

def createStringIO():
    if PY3K:
        from io import StringIO
    else:
        from StringIO import StringIO
    return StringIO()

def dictElement(element, prefix=None, number_multiples=False):
    """Returns a dictionary built from the children of *element*, which must be
    a :class:`xml.etree.ElementTree.Element` instance. Keys of the dictionary
    are *tag* of children without the *prefix*, or namespace. Values depend on
    the content of the child. If a child does not have any children, its text
    attribute is the value. If a child has children, then the child is the
    value.
    """
    
    dict_ = {}
    length = False
    if isinstance(prefix, str):
        length = len(prefix)

    prev_tag = ''
    for child in element:
        tag = child.tag

        if length and tag.startswith(prefix):
            tag = tag[length:]

        if tag != prev_tag:
            prev_tag = tag
            i = 0
        else:
            i += 1

        if number_multiples:
            tag = tag + '{:>4}'.format(str(i))
            
        if len(child) == 0:
            if child.text is None:
                dict_[tag] = child.items()
            else:
                dict_[tag] = child.text
        else:
            dict_[tag] = child

    return dict_

def dictElementLoop(dict_, keys=None, prefix=None, number_multiples=False):

    if isinstance(keys, str):
        keys = [keys]

    if not keys:
        keys = dict_.keys()

    for orig_key in keys:
        item = dict_[orig_key]
        if isinstance(item, Element):
            dict2 = dictElement(dict_[orig_key], prefix, number_multiples)
            finished = False
            while not finished:
                dict3 = dict2.copy()
                try:
                    key = dict2.keys()[0]
                    dict2[key] = dictElement(dict2[key], prefix, number_multiples)
                except:
                    finished = True
                else:
                    dict2 = dict3
                    for key in dict2.keys():
                        dict2[key] = dictElement(dict2[key], prefix, number_multiples)

            dict_[orig_key] = dict2

    return dict_

def intorfloat(x):
    """Returns ``int(x)``, or ``float(x)`` upon :exc:`ValueError`."""

    try:
        return int(x)
    except ValueError:
        return float(x)


def startswith(this, that):
    """Returns **True** if *this* or *that* starts with the other."""

    if len(this) < len(that):
        return that.startswith(this)
    else:
        return this.startswith(that)


def showFigure():
    """Call :func:`~matplotlib.pyplot.show` function with ``block=False``
    argument to avoid blocking behavior in non-interactive sessions.  If
    *block* keyword argument is not recognized, try again without it."""

    from matplotlib.pyplot import show
    try:
        show(block=False)
    except TypeError:
        show()


def countBytes(arrays, base=False):
    """Returns total number of bytes consumed by elements of arrays.  If
    *base* is **True**, use number of bytes from the base array."""

    if base:
        getbase = lambda arr: arr if arr.base is None else getbase(arr.base)
        nbytes = lambda arr: getbase(arr).nbytes
    else:
        nbytes = lambda arr: arr.nbytes

    return sum(nbytes(arr) for arr in arrays)

def sqrtm(matrix):
    """Returns the square root of a matrix."""
    (U,S,VT)=linalg.svd(matrix)
    D = diag(sqrt(S))
    return dot(dot(U,D),VT)

def getMasses(elements):
    """Gets the mass atom. """
    
    # mass_dict = {'C':12,'N':14,'S':32,'O':16,'H':1}
    mass_dict = IUPACData.atom_weights

    if isinstance(elements, str):
        return mass_dict[elements.capitalize()]
    else:
        masses = zeros(len(elements))
        for i,element in enumerate(elements):
            if element.capitalize() in mass_dict:
                masses[i] = mass_dict[element.capitalize()]
            else:
                masses[i] = 0.
        return masses

def count(L, a=None):
    return len([b for b in L if b is a])

def addEnds(x, y, axis=0):
    """Finds breaks in *x*, extends them by one position and adds **nan** at the 
    corresponding position in *y*. *x* needs to be an 1-D array, *y* can be a 
    matrix of column (or row) vectors"""

    d = diff(x)
    counter = Counter(d)
    step = counter.most_common(1)[0][0]

    breaks = where(d != step)[0]
    for b in reversed(breaks):
        x = insert(x, b+1, x[b]+step)
        y = insert(y, b+1, nan, axis=axis)

    return x, y

def copy(x):
    if x is None:
        return None
    elif isinstance(x, list):
        x = list(x)
    else:
        try:
            x = x.copy()
        except AttributeError:
            from copy import copy as shallow_copy
            
            x = shallow_copy(x)
    return x

def pystr(a):
    b = a
    if PY3K:
        if hasattr(a, 'decode'):
            b = a.decode() 
    else:
        if hasattr(a, 'encode'):
            b = a.encode() 
    return b

def getDataPath(filename):
    if PY3K:
        import importlib
        path = importlib.util.find_spec('prody.utilities.datafiles').submodule_search_locations[0]
        return '%s/%s' %(path, filename)
    import pkg_resources
    return pkg_resources.resource_filename('prody.utilities', 'datafiles/%s'%filename)      

def openData(filename, mode='r'):
    return open(getDataPath(filename), mode)

def saxsWater():
    filename = getDataPath('saxs_water.dat')
    return loadtxt(filename, delimiter=',')

def chr2(a):
    try:
        c = chr(a)
    except TypeError:
        c = str(a)
    return c

def toChararray(arr, aligned=False):
    arr = array(arr, dtype='|S')
    try:
        ndim, dtype_, shape = arr.ndim, arr.dtype, arr.shape
    except AttributeError:
        raise TypeError('arr is not a Numpy array')

    if ndim < 1:
        raise ValueError('arr.ndim should be at least 1')
    if dtype_.char != 'S':
        raise ValueError('arr must be a character array')

    if ndim != 2:
        n_seq = shape[0]
        l_seq = dtype_.itemsize
        new_arr = empty((n_seq, l_seq), dtype=bytes_)
        for i, s in enumerate(arr):
            for j in range(l_seq):
                if j < len(s):
                    new_arr[i, j] = chr2(s[j])
                else:
                    if aligned:
                        raise ValueError('arr does not the same lengths')
                    new_arr[i, j] = '.'
    else:
        new_arr = array(arr, dtype='|S1')
    return new_arr

def interpY(Y):
    Y = asarray(Y, dtype=float)
    n = len(Y)
    X = arange(n)

    dy = (Y.max() - Y.min()) / n

    Xp = [X[0]]; Yp = [Y[0]]
    for i in range(n-1):
        y1, y2 = Y[i], Y[i+1]
        x1, x2 = X[i], X[i+1]
        if abs(y2 - y1) > dy:
            sdy = sign(y2 - y1)*dy
            yp = arange(Y[i]+sdy, Y[i+1], sdy)
            xp = (yp - y1)/(y2 - y1)*(x2 - x1) + x1
            Xp.extend(xp)
            Yp.extend(yp)

        Xp.append(x2)
        Yp.append(y2)
    return array(Xp), array(Yp)

def cmp(a, b):
    return (a > b) - (a < b)

def getValue(dict_, attr, default=None):
    value = default
    if attr in dict_:
        value = dict_[attr]
        if default is not None:
            try:
                if value.ndim == 0:
                    value = type(default)(value)
            except:
                pass
    return value

def indentElement(elem, level=0):
    i = "\n" + level*"  "
    j = "\n" + (level-1)*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for subelem in elem:
            indentElement(subelem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = j
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = j
    return elem 

def isListLike(a):
    return isinstance(a, (list, tuple, ndarray))

def getDistance(coords1, coords2, unitcell=None):

    diff = coords1 - coords2
    if unitcell is not None:
        diff = subtract(diff, round(diff/unitcell)*unitcell, diff)
    return sqrt(power(diff, 2, diff).sum(axis=-1))

def fastin(a, B):
    for b in reversed(B):
        if a is b:
            return True
    return False

def div0(a, b, defval=0.):
    """ Performs ``true_divide`` but ignores the error when division by zero 
    (result is set to zero instead). """

    from numpy import errstate, true_divide, isfinite, isscalar
    
    with errstate(divide='ignore', invalid='ignore'):
        c = true_divide(a, b)
        if isscalar(c):
            if not isfinite(c):
                c = defval
        else:
            c[~isfinite(c)] = defval  # -inf inf NaN
    return c

def wmean(array, weights, axis=None):
    """Calculates the weighted average of *array* given *axis*."""

    try:
        avg = average(array, axis=axis, weights=weights)
    except ZeroDivisionError:
        numer = sum(array*weights, axis=axis)
        denom = sum(weights, axis=axis)
        avg = div0(numer, denom, nan)
    return avg

def bin2dec(x):
    """Converts the binary array to decimal."""

    y = 0
    for i,j in enumerate(x):
        if j: y += 1<<i
    return y


def wrapModes(modes):
    if hasattr(modes, 'getArray'):
        try:
            modes = [mode for mode in modes]
        except TypeError:
            modes = [modes]
    else:
        if isscalar(modes[0]):
            modes = [modes]
    return modes

def fixArraySize(arr, sizes, value=0):
    """Makes sure that **arr** is of **sizes**. If not, pad with **value**."""

    if not isinstance(arr, ndarray):
        raise TypeError('arr has to be a numpy ndarray')

    if not isinstance(sizes, tuple):
        raise TypeError('sizes has to be a tuple')

    shapes = arr.shape
    if shapes == sizes:
        return arr

    arr2 = empty(sizes, dtype=arr.dtype)
    arr2.fill(value)

    common_sizes = [min((a, b)) for a, b in zip(sizes, shapes)]
    slices = tuple(slice(s) for s in common_sizes)
    arr2[slices] = arr[slices]

    return arr2

def isSymmetric(M, rtol=1e-05, atol=1e-08):
    """Checks if the matrix is symmetric."""
    if M.shape != M.T.shape:
        return False
    return allclose(M, M.T, rtol=rtol, atol=atol)

def makeSymmetric(M):
    """Makes sure the matrix is symmetric."""
    
    if isSymmetric(M):
        return M

    # make square 
    n, m = M.shape
    l = max((n, m))
    M = fixArraySize(M, (l, l))

    # determine which part of the matrix has values
    U = triu(M, k=1)
    L = tril(M, k=-1)

    if U.sum() == 0:
        M += L.T
    elif L.sum() == 0:
        M += U.T
    else:
        M = (M + M.T) / 2.
    return M

def index(A, a):
    if isinstance(A, list):
        return A.index(a)
    else:
        A = asarray(A)
        return where(A==a)[0][0]


def checkIdentifiers(*pdb, **kwargs):
    """Check whether *pdb* identifiers are valid, and replace invalid ones
    with **None** in place."""
    format = kwargs.get('format', 'pdb')
    identifiers = []
    append = identifiers.append
    for pid in pdb:
        try:
            pid = pid.strip().lower()
        except AttributeError:
            LOGGER.warn('{0} is not a valid identifier.'.format(repr(pid)))
            append(None)
        else:
            if format != 'emd' and not ((len(pid) == 4 and pid.isalnum()) 
                                        or (len(pid) == 12 and pid.startswith('pdb_') 
                                            and pid[3] == '_')):
                LOGGER.warn('{0} is not a valid identifier.'
                            .format(repr(pid)))
                append(None)
            elif format == 'emd' and not (len(pid) in [4, 5] and pid.isalnum()):
                LOGGER.warn('{0} is not a valid identifier.'
                            .format(repr(pid)))
                append(None)
            else:
                append(pid)
    return identifiers

def decToBase36(integer):
    """Converts a decimal number to base 36.
    Based on https://wikivisually.com/wiki/Base36
    """

    chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    
    sign = '-' if integer < 0 else ''
    integer = abs(integer)
    result = ''
    
    while integer > 0:
        integer, remainder = divmod(integer, 36)
        result = chars[remainder]+result

    if result == '':
        result = '0'

    return sign+result


def decToHybrid36(x, resnum=False):
    """Convert a regular decimal number to a string in hybrid36 format"""
    if not isinstance(x, numbers.Integral):
        raise TypeError('x should be an integer')

    if not resnum:
        if x < 100000:
            return str(x)

        start = 10*36**4 # decToBase36(start) = A0000
        return decToBase36(int(x) + (start - 100000))
    else:
        if x < 10000:
            return str(x)

        start = 10*36**3 # decToBase36(start) = A000
        return decToBase36(int(x) + (start - 10000))        


def base36ToDec(x):
    chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    if x[0] == '-':
        sign = '-'
        x = x[1:]
    else:
        sign = ''

    result = 0
    for i, entry in enumerate(reversed(x)):
        result += (chars.find(entry)*36**(i))

    return int(sign + str(result))


def hybrid36ToDec(x, resnum=False):
    """Convert string in hybrid36 format to a regular decimal number"""
    if not isinstance(x, str):
        raise TypeError('x should be a string')
    
    isnumeric = all([y.isdigit() for y in x])
    if isnumeric:
        return int(x)

    if not resnum:
        start = 10*36**4 # decToBase36(start) = A0000
        return base36ToDec(x) - start + 100000
    else:
        start = 10*36**3 # decToBase36(start) = A000
        return base36ToDec(x) - start + 10000        


def split(string, shlex=False):
    if shlex:
        import shlex
        return shlex.split(string)
    else:
        return string.split()


def packmolRenumChains(ag):
    chainids = ag.getChids()
    new_chainids = zeros(chainids.shape[0], dtype=DTYPE + '2')
    chid_mem = []
    chid_alt = 0

    num_chars = int(chainids.dtype.str[2:])
    for j, chid in enumerate(chainids):
        if j == 0:
            chid_mem.append(chid)
        else:
            if chid != chainids[j-1]:
                chid_alt += 1

                if chid_alt >= 10**num_chars:
                    num_chars += 1
                    new_chainids = array(new_chainids, dtype=new_chainids.dtype.str[:2]+str(num_chars))

                if chid_alt > 1 and len(chid_mem) > 1 and chid == chid_mem[-2]:
                    chid_mem.append(chid)

        new_chainids[j] = "{}".format(chid_alt)

    ag.setChids(new_chainids)
    return ag

def mad(x):
    try:
        from scipy.stats import median_absolute_deviation as _mad
    except ImportError:
        try:
            from scipy.stats import median_abs_deviation as _mad
        except ImportError:
            def _mad(x):
                med = median(x)
                return median(abs(x - med))
    
    return _mad(x)


def importDec():
    """Returns an imported module equivalent to numpy testing decorators."""

    try:
        import numpy.testing.decorators as dec
    except ModuleNotFoundError:
        try:
            from numpy.testing import dec
        except ImportError:
            try:
                import numpy.testing._private.decorators as dec
            except ModuleNotFoundError:
                from pytest import mark as dec    

    return dec


def impLoadModule(name, cmd, path):
    """Returns an  an imported module equivalent to imp."""

    if not name.endswith('.'):
        name += '.'

    try:
        import imp
        mod = imp.load_module(name + cmd,
                              *imp.find_module(cmd, [path]))
    except ImportError:
        import importlib
        loader = importlib.machinery.SourceFileLoader(name + cmd,
                                                      path + cmd + '.py')
        mod = importlib.util.types.ModuleType(loader.name)
        loader.exec_module(mod)

    return mod



# -*- coding: utf-8 -*-
"""This module defines classes for handling trajectory files in DCD format.


Parse/write DCD files
===============================================================================

  * :class:`.DCDFile`
  * :func:`.parseDCD`
  * :func:`.writeDCD`

Parse structure files
===============================================================================

  * :func:`.parsePSF`

Handle multiple files
===============================================================================

  * :class:`.Trajectory`

Handle frame data
===============================================================================

  * :class:`.Frame`

Examples
===============================================================================

Following examples show how to use trajectory classes and functions:

  * :ref:`trajectory`
  * :ref:`trajectory2`
  * :ref:`eda`"""

from os.path import splitext

__all__ = []

def openTrajFile(filename, *args, **kwargs):

    ext = splitext(filename)[1][1:].lower()
    try:
        return TRAJFILE[ext](filename, *args, **kwargs)
    except KeyError:
        raise ValueError('trajectory file type {0} is not recognized'
                         .format(repr(ext)))

from . import trajbase
from .trajbase import *
__all__.extend(trajbase.__all__)

from . import trajectory
from .trajectory import *
__all__.extend(trajectory.__all__)

from . import dcdfile
from .dcdfile import *
__all__.extend(dcdfile.__all__)

from . import frame
from .frame import *
__all__.extend(frame.__all__)

from . import psffile
from .psffile import *
__all__.extend(psffile.__all__)

TRAJFILE = {'dcd': DCDFile}




'''
Copyright (c) 2020 Burak Kaynak, Pemra Doruker.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

__author__ = 'Burak Kaynak'
__credits__ = 'Pemra Doruker'
__email__ = ['burak.kaynak@pitt.edu', 'doruker@pitt.edu']

from collections import defaultdict
from os import chdir, listdir, mkdir, system
from os.path import isdir
from numpy import argsort, arange, array, c_, count_nonzero, hstack, mean, median, quantile, save, where
from scipy.stats import zscore

from prody import LOGGER
from prody.atomic.functions import extendAtomicData
from .anm import ANM
from .gnm import GNM
from prody.proteins import parsePDB, writePDB
from .editing import reduceModel
from .plotting import showAtomicLines
from .signature import ModeEnsemble, saveModeEnsemble
from prody.utilities import which, isListLike
from . import matchModes

__all__ = ['ESSA']


class ESSA:

    '''
    ESSA determines the essentiality score of each residue based on the extent to which it can alter the global dynamics ([KB20]_). It can also rank potentially allosteric pockets by calculating their ESSA and local hydrophibic density z-scores using Fpocket algorithm ([LGV09]_). 

    .. [KB20] Kaynak B.T., Bahar I., Doruker P., Essential site scanning analysis: A new approach for detecting sites that modulate the dispersion of protein global motions, *Comput. Struct. Biotechnol. J.* **2020** 18:1577-1586.

    .. [LGV09] Le Guilloux, V., Schmidtke P., Tuffery P., Fpocket: An open source platform for ligand pocket detection, *BMC Bioinformatics* **2009** 10:168.

    Instantiate an ESSA object.
    '''

    _single = {'GLY': 'G', 'ALA': 'A', 'LEU': 'L', 'MET': 'M',
               'PHE': 'F', 'TRP': 'W', 'LYS': 'K', 'GLN': 'Q',
               'GLU': 'E', 'SER': 'S', 'PRO': 'P', 'VAL': 'V',
               'ILE': 'I', 'CYS': 'C', 'TYR': 'Y', 'HIS': 'H',
               'ARG': 'R', 'ASN': 'N', 'ASP': 'D', 'THR': 'T'}
    
    def __init__(self):

        self._atoms = None
        self._title = None
        self._heavy = None
        self._ca = None
        self._n_modes = None
        self._enm = None
        self._cutoff = None
        self._lig = None
        self._ligres_idx = None
        self._ligres_code = None
        self._rib = None
        self._ri = None
        self._chrn = None
        self._dist = None
        self._ensemble = None
        self._labels = None
        self._zscore = None
        self._ref = None
        self._eigvals = None
        self._eigvecs = None

    def setSystem(self, atoms, **kwargs):

        '''
        Sets atoms, ligands and a cutoff distance for protein-ligand interactions.

        :arg atoms: *atoms* parsed by parsePDB

        :arg lig: String of ligands' chainIDs and resSeqs (resnum) separated by a whitespace,
            e.g., 'A 300 B 301'. Default is None.
        :type lig: str

        :arg dist: Atom-atom distance (A) to select the protein residues that are in contact with a ligand, default is 4.5 A.
        :type dist: float

        :arg lowmem: If True, a ModeEnsemble is not generated due to the lack of memory resources, and eigenvalue/eigenvectors are only stored, default it False.
        :type lowmem: bool
        '''

        self._atoms = atoms
        self._title = atoms.getTitle()
        self._lig = kwargs.pop('lig', None)
        if self._lig:
            self._ligres_idx = {}
            self._ligres_code = {}
            self._dist = kwargs.pop('dist', 4.5)

        self._heavy = atoms.select('protein and heavy and not hetatm')
        self._ca = self._heavy.ca

        self._lowmem = kwargs.pop('lowmem', False)
        if self._lowmem:
            self._eigvals = []
            self._eigvecs = []

        self._rib = all(self._ca.getResindices() == arange(self._ca.numAtoms()))
        if not self._rib:
            self._ri = {v: k for k, v in enumerate(self._ca.getResindices())}

        # --- resindices of protein residues that are within dist A of ligands --- #

        if self._lig:
            ligs = self._lig.split()
            ligs = list(zip(ligs[::2], ligs[1::2]))
            for chid, resnum in ligs:
                key = ''.join(chid + str(resnum))
                sel_lig = 'calpha and not hetatm and (same residue as ' \
                          'exwithin {} of (chain {} and resnum {}))'.format(self._dist, chid, resnum)
                self._ligres_idx[key] = self._atoms.select(sel_lig).getResindices()

            for k, v in self._ligres_idx.items():
                atoms = self._ca.select('resindex ' + ' '.join([str(i) for i in v]))
                tmp0 = defaultdict(list)
                for ch, rn in zip(atoms.getChids(), atoms.getResnums()):
                    tmp0[ch].append(str(rn))
                tmp1 = {ch: ' '.join(rn) for ch, rn in tmp0.items()}
                self._ligres_code[k] = ['chain {} and resnum {}'.format(ch, rn) for ch, rn in tmp1.items()]

    def scanResidues(self, n_modes=10, enm='gnm', cutoff=None):

        '''
        Scans residues to generate ESSA z-scores.

        :arg n_modes: Number of global modes.
        :type n_modes: int

        :arg enm: Type of elastic network model, 'gnm' or 'anm', default is 'gnm'.
        :type enm: str

        :arg cutoff: Cutoff distance (A) for pairwise interactions, default is 10 A for GNM and 15 A for ANM.
        :type cutoff: float
        '''

        self._n_modes = n_modes
        self._enm = enm
        self._cutoff = cutoff

        self._ensemble = ModeEnsemble('{}'.format(self._title))
        self._ensemble.setAtoms(self._ca)
        self._labels = ['ref']

        # --- reference model --- #

        self._reference()

        # --- perturbed models --- #

        LOGGER.progress(msg='', steps=(self._ca.numAtoms()))
        for i, j in enumerate(self._ca.getResindices()):
            LOGGER.update(step=i+1, msg='scanning residue {}'.format(i+1))
            self._perturbed(j)

        if self._lowmem:
            self._eigvals = array(self._eigvals)
            self._eigvecs = array(self._eigvecs)

        # --- ESSA computation part --- #

        if self._lowmem:
            denom = self._eigvals[0]
            num = self._eigvals[1:] - denom
        else:
            self._ensemble.setLabels(self._labels)
            self._ensemble.match()

            denom = self._ensemble[0].getEigvals()
            num = self._ensemble[1:].getEigvals() - denom

        eig_diff = num / denom * 100
        eig_diff_mean = mean(eig_diff, axis=1)

        self._zscore = zscore(eig_diff_mean)

        if self._lig:
            self._zs_lig = {}
            for k, v in self._ligres_idx.items():
                if self._rib:
                    self._zs_lig[k] = [array(v), self._zscore[v]]
                else:
                    vv = [self._ri[i] for i in v]
                    self._zs_lig[k] = [array(vv), self._zscore[vv]]

    def _reference(self):

        if self._enm == 'gnm':
            ca_enm = GNM('ca')
            if self._cutoff is not None:
                ca_enm.buildKirchhoff(self._ca, cutoff=self._cutoff)
            else:
                ca_enm.buildKirchhoff(self._ca)
                self._cutoff = ca_enm.getCutoff()

        if self._enm == 'anm':
            ca_enm = ANM('ca')
            if self._cutoff is not None:
                ca_enm.buildHessian(self._ca, cutoff=self._cutoff)
            else:
                ca_enm.buildHessian(self._ca)
                self._cutoff = ca_enm.getCutoff()

        ca_enm.calcModes(n_modes=self._n_modes)

        if self._lowmem:
            self._ref = ca_enm
            self._eigvals.append(ca_enm.getEigvals())
            self._eigvecs.append(ca_enm.getEigvecs())
        else:
            self._ensemble.addModeSet(ca_enm[:])

    def _perturbed(self, arg):

        sel = 'calpha or resindex {}'.format(arg)
        tmp = self._heavy.select(sel)

        if self._enm == 'gnm':
            tmp_enm = GNM('res_{}'.format(arg))
            tmp_enm.buildKirchhoff(tmp, cutoff=self._cutoff)

        if self._enm == 'anm':
            tmp_enm = ANM('res_{}'.format(arg))
            tmp_enm.buildHessian(tmp, cutoff=self._cutoff)

        tmp_enm_red, _ = reduceModel(tmp_enm, tmp, self._ca)
        tmp_enm_red.calcModes(n_modes=self._n_modes)
        tmp_enm_red.setTitle(tmp_enm_red.getTitle().split()[0])

        if self._lowmem:
            _, matched = matchModes(self._ref, tmp_enm_red)
            self._eigvals.append(matched.getEigvals())
            self._eigvecs.append(matched.getEigvecs())
        else:
            self._ensemble.addModeSet(tmp_enm_red[:])

        self._labels.append(tmp_enm_red.getTitle())

    def getESSAZscores(self):

        'Returns ESSA z-scores.'

        return self._zscore

    def getESSAEnsemble(self):

        'Returns ESSA mode ensemble, comprised of ENMS calculated for each scanned/perturbed residue.'

        if self._lowmem:
            LOGGER.warn('ModeEnsemble was not generated due to lowmem=True')
        else:
            return self._ensemble[:]
    
    def saveESSAEnsemble(self):

        'Saves ESSA mode ensemble, comprised of ENMS calculated for each scanned/perturbed residue.'

        if self._lowmem:
            LOGGER.warn('ModeEnsemble was not generated due to lowmem=True')
        else:
            saveModeEnsemble(self._ensemble, filename='{}_{}'.format(self._title, self._enm))

    def saveESSAZscores(self):

        'Saves ESSA z-scores to a binary file in Numpy `.npy` format.'

        save('{}_{}_zs'.format(self._title, self._enm), self._zscore)

    def writeESSAZscoresToPDB(self):

        'Writes a pdb file with ESSA z-scores placed in the B-factor column.'

        writePDB('{}_{}_zs'.format(self._title, self._enm), self._heavy,
                 beta=extendAtomicData(self._zscore, self._ca, self._heavy)[0])

    def getEigvals(self):

        'Returns eigenvalues of the matched modes.'

        if self._lowmem:
            return self._eigvals
        else:
            return self._ensemble.getEigvals()

    def getEigvecs(self):

        'Returns eigenvectors of the matched modes.'

        if self._lowmem:
            return self._eigvecs
        else:
            return self._ensemble.getEigvecs().getArray()

    def saveEigvals(self):

        'Saves eigenvalues of the matched modes in Numpy `.npy` format.'

        if self._lowmem:
            save('{}_{}_eigvals'.format(self._title, self._enm), self._eigvals)
        else:
            save('{}_{}_eigvals'.format(self._title, self._enm), self._ensemble.getEigvals())

    def saveEigvecs(self):

        'Saves eigenvectors of the matched modes in Numpy `.npy` format.'

        if self._lowmem:
            save('{}_{}_eigvecs'.format(self._title, self._enm), self._eigvecs)
        else:
            save('{}_{}_eigvecs'.format(self._title, self._enm), self._ensemble.getEigvecs().getArray())

    def getLigandResidueESSAZscores(self):

        'Returns ESSA z-scores of the residues interacting with ligands as a dictionary. The keys of which are the corresponding chain ids and residue numbers of the ligands. Each value comprises the indices of the residue ESSA z-scores in the profile and the corresponding scores as separate arrays.'
        
        if self._lig:
            return self._zs_lig
        else:
            LOGGER.warning('No ligand provided.')

    def saveLigandResidueESSAZscores(self):

        'Saves the dictionary of ESSA z-scores of the residues interacting with ligands to a pickle `.pkl` file. The keys of the dictionary are the corresponding chain ids and residue numbers of the ligands. Each value comprises the indices of the residue ESSA z-scores in the profile and the corresponding scores as separate arrays.'

        if self._lig:
            from pickle import dump

            dump(self._zs_lig, open('{}_ligres_gnm_zs.pkl'.format(self._title), 'wb'))
        else:
            LOGGER.warning('No ligand provided.')
        
    def getLigandResidueIndices(self):

        'Returns residue indices of the residues interacting with ligands.'

        if self._lig:
            return self._ligres_idx
        else:
            LOGGER.warning('No ligand provided.')

    def getLigandResidueCodes(self):

        'Returns chain ids and residue numbers of the residues interacting with ligands.'

        if self._lig:
            return self._ligres_code
        else:
            LOGGER.warning('No ligand provided.')

    def saveLigandResidueCodes(self):

        'Saves chain ids and residue numbers of the residues interacting with ligands.'

        if self._lig:
            with open('{}_ligand_rescodes.txt'.format(self._title), 'w') as f:
                for k, v in self._ligres_code.items():
                    f.write(k + '\n')
                    for x in v:
                        f.write(x + '\n')
        else:
            LOGGER.warning('No ligand provided.')

    def _codes(self, arg):

        sel = self._ca.select('resindex {}'.format(arg))
        try:
            return self._single[sel.getResnames()[0]] + str(sel.getResnums()[0])
        except KeyError:
            return sel.getResnames()[0] + str(sel.getResnums()[0])

    def showESSAProfile(self, q=.75, rescode=False, sel=None):

        '''
        Shows ESSA profile.

        :arg q: Quantile value to plot a baseline for z-scores, default is 0.75. If it is set to 0.0, then the baseline is not drawn.
        :type q: float

        :arg rescode: If it is True, the ligand interacting residues with ESSA scores larger than the baseline are denoted by their single letter codes and chain ids on the profile. If quantile vaue is 0.0, then all ligand-interacting residues denoted by their codes.
        :type rescode: bool

        :arg sel: It is a selection string, default is None. If it is provided, then selected residues are shown with their single letter codes and residue numbers on the profile. For example, 'chain A and resnum 33 47'.
        :type sel: str
        '''

        from matplotlib.pyplot import text, legend, hlines, scatter, xlabel, ylabel, tight_layout

        showAtomicLines(self._zscore, atoms=self._ca, c='k', linewidth=1.)

        if self._lig:
            for k in self._zs_lig.keys():
                scatter(*self._zs_lig[k], label=k)
                if rescode:
                    if q != 0.0:
                        idx = where(self._zs_lig[k][1] >= quantile(self._zscore, q=q))[0]
                        if idx.size > 0:
                            _x = self._zs_lig[k][0][idx]
                            _y = self._zs_lig[k][1][idx]
                            _i = self._ligres_idx[k][idx]
                        else:
                            break
                    else:
                        _x = self._zs_lig[k][0]
                        _y = self._zs_lig[k][1]
                        _i = self._ligres_idx[k]
                    for x, y, i in zip(_x, _y, _i):
                        text(x, y, self._codes(i), color='r')
            legend()

        if q != 0.0:
            hlines(quantile(self._zscore, q=q),
                       xmin=0., xmax=self._ca.numAtoms(),
                       linestyle='--', color='c')

        if sel:
            idx = self._ca.select(sel).getResindices()
            if self._rib:
                _x = idx
                zs_sel = self._zscore[_x]
            else:
                _x = [self._ri[i] for i in idx]
                zs_sel = self._zscore[_x]

            scatter(_x, zs_sel)
            for x, y, i in zip(_x, zs_sel, idx):
                text(x, y, self._codes(i), color='r')
            

        xlabel('Residue')
        ylabel('Z-Score')

        tight_layout()

    def scanPockets(self):

        'Generates ESSA z-scores for pockets and parses pocket features. It requires both Fpocket 3.0 and Pandas being installed in your system.'
        
        from re import findall

        fpocket = which('fpocket')

        if fpocket is None:
            LOGGER.warning('Fpocket (version >= 3.0) was not found, please install it.')
            return None

        try:
            from pandas import Index, DataFrame
        except ImportError as ie:
            LOGGER.warning(ie.__str__() + ' was found, please install it.')
            return None

        rcr = {(i, j): k if self._rib else self._ri[k]
               for i, j, k in zip(self._ca.getChids(),
                                  self._ca.getResnums(),
                                  self._ca.getResindices())}

        writePDB('{}_pro'.format(self._title), self._heavy)

        direc = '{}_pro_out'.format(self._title)
        if not isdir(direc):
            system('fpocket -f {}_pro.pdb'.format(self._title))

        chdir(direc + '/pockets')
        l = [x for x in listdir('.') if x.endswith('.pdb')]
        l.sort(key=lambda x:int(x.partition('_')[0][6:]))

        ps = []
        for x in l:
            with open(x, 'r') as f:
                tmp0 = f.read()
                tmp1 = [(x[1].strip(), float(x[2])) for x in findall(r'(\w+\s\w+\s*-\s*)(.+):\s*([\d.-]+)(\n)', tmp0)]
            fea, sco = list(zip(*tmp1))
            ps.append(sco)
        pdbs = parsePDB(l)
        if not isListLike(pdbs):
            pdbs = [pdbs]
        chdir('../..')

        # ----- # ----- #

        ps = array(ps)

        pcn = {int(pdb.getTitle().partition('_')[0][6:]):
               set(zip(pdb.getChids().tolist(),
                       pdb.getResnums().tolist())) for pdb in pdbs}
        pi = {p: [rcr[x] for x in crn] for p, crn in pcn.items()}

        pzs_max = {k: max(self._zscore[v]) for k, v in pi.items()}
        pzs_med = {k: median(self._zscore[v]) for k, v in pi.items()}

        # ----- # ----- #

        indices = Index(range(1, ps.shape[0] + 1), name='Pocket #')

        columns = Index(fea, name='Feature')

        self._df = DataFrame(index=indices, columns=columns, data=ps)

        # ----- # ----- #

        columns_zs = Index(['ESSA_max',
                            'ESSA_med',
                            'LHD'],
                           name='Z-score')

        zps = c_[list(pzs_max.values())]
        zps = hstack((zps, c_[list(pzs_med.values())]))
        zps = hstack((zps, zscore(self._df[['Local hydrophobic density Score']])))


        self._df_zs = DataFrame(index=indices, columns=columns_zs, data=zps)

    def rankPockets(self):

        'Ranks pockets in terms of their allosteric potential, based on their ESSA z-scores (max/median) with local hydrophobic density (LHD) screening.'

        from pandas import DataFrame, Index

        lhd = self._df_zs.loc[:, 'LHD']
        n = count_nonzero(lhd >= 0.)
        q = quantile(lhd, q=.85)

        # ----- # ------ #

        s_max = ['ESSA_max', 'LHD']

        zf_max = self._df_zs[s_max].copy()

        if n >= lhd.size // 4:
            f_max = zf_max.iloc[:, 1] >= 0.
        else:
            f_max = zf_max.iloc[:, 1] >= q

        zf_max = zf_max[f_max]

        zf_max.iloc[:, 0] = zf_max.iloc[:, 0].round(1)
        zf_max.iloc[:, 1] = zf_max.iloc[:, 1].round(2)

        self._idx_max = zf_max.sort_values(s_max, ascending=False).index

        # ----- # ----- #

        s_med = ['ESSA_med',
                 'LHD']

        zf_med = self._df_zs[s_med].copy()

        if n >= lhd.size // 4:
            f_med = zf_med.iloc[:, 1] >= 0.
        else:
            f_med = zf_med.iloc[:, 1] >= q

        zf_med = zf_med[f_med]

        zf_med.iloc[:, 0] = zf_med.iloc[:, 0].round(1)
        zf_med.iloc[:, 1] = zf_med.iloc[:, 1].round(2)

        self._idx_med = zf_med.sort_values(s_med, ascending=False).index

        ranks = Index(range(1, zf_max.shape[0] + 1), name='Rank')

        columns_ranks = Index(['Pocket # (ESSA_max & LHD)', 'Pocket # (ESSA_med & LHD)'])

        self._pocket_ranks = DataFrame(index=ranks, columns=columns_ranks)

        self._pocket_ranks.iloc[:, 0] = self._idx_max
        self._pocket_ranks.iloc[:, 1] = self._idx_med

    def getPocketFeatures(self):

        'Returns pocket features as a Pandas dataframe.'

        return self._df

    def getPocketZscores(self):

        'Returns ESSA and local hydrophobic density (LHD) z-scores for pockets as a Pandas dataframe.'

        return self._df_zs

    def getPocketRanks(self):

        'Returns pocket ranks (allosteric potential).'

        return self._pocket_ranks
    
    def showPocketZscores(self):

        'Plots maximum/median ESSA and local hydrophobic density (LHD) z-scores for pockets.'

        from matplotlib.pyplot import style, xticks, xlabel, ylabel, tight_layout

        with style.context({'xtick.major.size': 10, 'xtick.labelsize': 30,
                                'ytick.major.size': 10, 'ytick.labelsize': 30,
                                'axes.labelsize': 35, 'legend.fontsize': 25,
                                'legend.title_fontsize': 0}):
            self._df_zs[['ESSA_max',
                         'ESSA_med',
                         'LHD']].plot.bar(figsize=(25, 10))
            xticks(rotation=0)
            xlabel('Pocket')
            ylabel('Z-score')
            tight_layout()

    def savePocketFeatures(self):

        'Saves pocket features to a pickle `.pkl` file.'

        self._df.to_pickle('{}_pocket_features.pkl'.format(self._title))

    def savePocketZscores(self):

        'Saves ESSA and local hydrophobic density (LHD) z-scores of pockets to a pickle `.pkl` file.'

        self._df_zs.to_pickle('{}_pocket_zscores.pkl'.format(self._title))

    def savePocketRanks(self):

        'Saves pocket ranks to a binary file in Numpy `.npy` format.'

        save('{}_{}_pocket_ranks_wrt_ESSAmax_LHD'.format(self._title, self._enm),
             self._idx_max)
        save('{}_{}_pocket_ranks_wrt_ESSAmed_LHD'.format(self._title, self._enm),
             self._idx_med)
        
    def writePocketRanksToCSV(self):

        'Writes pocket ranks to a `.csv` file.'

        self._pocket_ranks.to_csv('{}_{}_pocket_ranks.csv'.format(self._title, self._enm), index=False)



"""This module defines functions for handling files and paths."""

import os
import sys
from os import sep as pathsep
from glob import glob as pyglob
import pickle as pypickle
import zipfile
import platform
import os.path
from os.path import isfile, isdir, join, split, splitext
from os.path import getsize, isabs, exists, abspath
from shutil import copy, Error as shError

from prody import PY3K

PLATFORM = platform.system()
USERHOME = os.getenv('USERPROFILE') or os.getenv('HOME')

__all__ = ['gunzip', 'backupFile', 'openFile',
           'openDB', 'openSQLite', 'openURL', 'copyFile',
           'isExecutable', 'isReadable', 'isWritable',
           'makePath', 'relpath', 'sympath', 'which',
           'pickle', 'unpickle', 'glob', 'addext',
           'PLATFORM', 'USERHOME']

if PY3K:
    import gzip
    from gzip import GzipFile
    import io

    class TextIOWrapper(io.TextIOWrapper):

        def _getlines(self):

            try:
                lines = self._lines
            except AttributeError:
                self._lines = None

            if self._lines is None:
                self._lines = self.read().split('\n')
            return self._lines

        def readline(self, *args):

            lines = self._getlines()
            if lines:
                return lines.pop(0)
            else:
                return ''

        def readlines(self, size=None):

            lines = self._getlines()
            if size is None:
                self._lines = []
                return lines
            else:
                self._lines = lines[size:]
                return lines[:size]

        def __del__(self):

            self.close()

    def gzip_open(filename, mode="rb", compresslevel=9,
                  encoding=None, errors=None, newline=None):
        """Open a gzip-compressed file in binary or text mode.

        The filename argument can be an actual filename (a str or bytes object), or
        an existing file object to read from or write to.

        The mode argument can be "r", "rb", "w", "wb", "a" or "ab" for binary mode,
        or "rt", "wt" or "at" for text mode. The default mode is "rb", and the
        default compresslevel is 9.

        For binary mode, this function is equivalent to the GzipFile constructor:
        GzipFile(filename, mode, compresslevel). In this case, the encoding, errors
        and newline arguments must not be provided.

        For text mode, a GzipFile object is created, and wrapped in an
        io.TextIOWrapper instance with the specified encoding, error handling
        behavior, and line ending(s).

        """
        if "t" in mode:
            if "b" in mode:
                raise ValueError("Invalid mode: %r" % (mode,))
        else:
            if encoding is not None:
                raise ValueError("Argument 'encoding' not supported in binary mode")
            if errors is not None:
                raise ValueError("Argument 'errors' not supported in binary mode")
            if newline is not None:
                raise ValueError("Argument 'newline' not supported in binary mode")

        gz_mode = mode.replace("t", "")
        if isinstance(filename, (str, bytes)):
            binary_file = GzipFile(filename, gz_mode, compresslevel)
        elif hasattr(filename, "read") or hasattr(filename, "write"):
            binary_file = GzipFile(None, gz_mode, compresslevel, filename)
        else:
            raise TypeError("filename must be a str or bytes object, or a file")

        if "t" in mode:
            return TextIOWrapper(binary_file, encoding, errors, newline)
        else:
            return binary_file
else:
    import gzip
    def gzip_open(filename, *args, **kwargs):
        if args and "t" in args[0]:
            args = (args[0].replace("t", ""), ) + args[1:]
        if isinstance(filename, str):
            return gzip.open(filename, *args, **kwargs)
        else:
            return gzip.GzipFile(filename, *args, **kwargs)

if PY3K:
    from gzip import compress as gzip_compress
    from gzip import decompress as gzip_decompress

OPEN = {
    '.gz': gzip_open,
    '.zip': zipfile.ZipFile,
}


def backupFile(filename, backup=None, backup_ext='.BAK', **kwargs):
    """Rename *filename* with *backup_ext* appended to its name for backup
    purposes, if *backup* is **True** or if automatic backups is turned on
    using :func:`.confProDy`.  Default extension :file:`.BAK` is used when
    one is not set using :func:`.confProDy`.  If *filename* does not exist,
    no action will be taken and *filename* will be returned.  If file is
    successfully renamed, new filename will be returned."""

    try:
        exists = isfile(filename)
    except Exception as err:
        raise TypeError('filename must be a string ({0})'.format(str(err)))

    from prody import SETTINGS
    if exists and (backup or SETTINGS.get('backup', False)):
        if backup_ext == '.BAK':
            backup_ext = SETTINGS.get('backup_ext', '.BAK')
        bak = filename + backup_ext
        if isfile(bak):
            try:
                os.remove(bak)
            except Exception as err:
                pass
        try:
            os.rename(filename, bak)
        except Exception as err:
            pass
        return bak
    else:
        return filename


def openFile(filename, *args, **kwargs):
    """Open *filename* for reading, writing, or appending.  First argument in
    *args* is treated as the mode.  Opening :file:`.gz` and :file:`.zip` files
    for reading and writing is handled automatically.

    :arg backup: backup existing file using :func:`.backupFile` when opening
        in append or write modes, default is obtained from package settings
    :type backup: bool

    :arg backup_ext: extension for backup file, default is :file:`.BAK`
    :type backup_ext: str"""

    from prody import SETTINGS
    try:
        exists = isfile(filename)
    except Exception as err:
        raise TypeError('filename must be a string ({0})'.format(str(err)))

    folder = kwargs.pop('folder', None)
    if folder:
        filename = join(folder, filename)

    backup = kwargs.pop('backup', None)
    if backup is not None and backup and args and args[0][0] in ('a', 'w'):
        backupFile(filename, backup=backup,
                   backup_ext=kwargs.pop('backup_ext', None))

    ext = splitext(filename)[1]
    return OPEN.get(ext.lower(), open)(filename, *args, **kwargs)


def gunzip(filename, outname=None):
    """Returns output name that contains decompressed contents of *filename*.
    When no *outname* is given, *filename* is used as the output name as it
    is or after :file:`.gz` extension is removed.  *filename* may also be a
    string buffer, in which case decompressed string buffer or *outname* that
    contains buffer will be returned."""

    if len(filename) < 1000:
        try:
            afile = isfile(filename)
        except (TypeError, ValueError):
            afile = False
    else:
        afile = False

    if afile:
        if outname is None:
            if filename.lower().endswith('.gz'):
                outname = filename[:-3]
            elif filename.lower().endswith('.tgz'):
                outname = filename[:-4] + '.tar'
            elif filename.lower().endswith('.gzip'):
                outname = filename[:-5]
            else:
                outname = filename

        inp = gzip_open(filename, 'rb')
        data = inp.read()
        inp.close()
        out = open(outname, 'wb')
        out.write(data)
        out.close()
        return outname
    else:
        result = None
        try:
            from StringIO import StringIO
        except (ImportError, ModuleNotFoundError):
            from io import BytesIO
            buff = gzip_open(BytesIO(filename))
            if outname is None:
                try:
                    result = buff.read()
                except IOError:
                    pass
            else:
                buff = buff.read()
                if isinstance(buff, bytes):
                    with open(outname, 'wb') as out:
                        out.write(buff)
                else:
                    with open(outname, 'wb') as out:
                        out.write(buff)
                return outname
        else:
            from StringIO import StringIO
            buff = gzip.GzipFile(fileobj=StringIO(filename))
            try:
                result = buff.read()
            except IOError:
                pass

        if result is not None:
            if outname is None:
                return result
            else:
                with open(outname, 'w') as out:
                    out.write(result)
                return outname
        raise ValueError('filename is not a valid path or a compressed'
                         ' string buffer')


def isExecutable(path):
    """Returns true if *path* is an executable."""

    return (isinstance(path, str) and exists(path) and
        os.access(path, os.X_OK))


def isReadable(path):
    """Returns true if *path* is readable by the user."""

    return (isinstance(path, str) and exists(path) and
        os.access(path, os.R_OK))


def isWritable(path):
    """Returns true if *path* is writable by the user."""

    return (isinstance(path, str) and exists(path) and
        os.access(path, os.W_OK))


def relpath(path):
    """Returns *path* on Windows, and relative path elsewhere."""

    if PLATFORM == 'Windows':
        return path
    else:
        return os.path.relpath(path)


def sympath(path, beg=2, end=1, ellipsis='...'):
    """Returns a symbolic path for a long *path*, by replacing folder names
    in the middle with *ellipsis*.  *beg* and *end* specified how many folder
    (or file) names to include from the beginning and end of the path."""

    abs_items = abspath(path).split(pathsep)
    rel_items = relpath(path).split(pathsep)
    if len(abs_items) <= len(rel_items):
        items = abs_items
    else:
        items = rel_items
    if len(items) <= beg + end:
        return pathsep.join(items)
    else:
        return pathsep.join(items[:beg+1] + [ellipsis] + items[-end:])


def makePath(path):
    """Make all directories that does not exist in a given *path*."""

    if not isdir(path):
        dirs = path.split(pathsep)
        for i, dirname in enumerate(dirs):
            if not dirname:
                continue
            dirname = pathsep.join(dirs[:i+1])
            try:
                if not isdir(dirname):
                    os.mkdir(dirname)
            except OSError:
                raise OSError('{0} could not be created, please '
                            'specify another path'.format(path))
    return path


def which(program):
    """This function is based on the example in:
    http://stackoverflow.com/questions/377017/"""

    fpath, fname = os.path.split(program)
    fname, fext = os.path.splitext(fname)

    if fpath and isExecutable(program):
        return program
    else:
        if os.name == 'nt' and fext == '':
            program += '.exe'
        for path in os.environ["PATH"].split(os.pathsep):
            path = os.path.join(path, program)
            if isExecutable(path):
                return path
    return None


def pickle(obj, filename, protocol=2, **kwargs):
    """Pickle *obj* using :func:`pickle.dump` in *filename*.  *protocol* is set
    to 2 for compatibility between Python 2 and 3."""

    out = openFile(filename, 'wb', **kwargs)
    pypickle.dump(obj, out, protocol=2)
    out.close()
    return filename


def unpickle(filename, **kwargs):
    """Unpickle object in *filename* using :func:`pickle.load`."""

    inf = openFile(filename, 'rb', **kwargs)
    obj = pypickle.load(inf)
    inf.close()
    return obj


def openDB(filename, *args):
    """Open a database with given *filename*."""

    try:
        import anydbm as dbm
    except ImportError:
        import dbm
    return dbm.open(filename, *args)


def openSQLite(filename, *args):
    """Returns a connection to SQLite database *filename*.  If ``'n'`` argument
    is passed, remove any existing databases with the same name and return
    connection to a new empty database."""

    if 'n' in args and isfile(filename):
        os.remove(filename)
    import sqlite3
    return sqlite3.connect(filename)


def openURL(url, timeout=5, **kwargs):
    """Open *url* for reading. Raise an :exc:`IOError` if *url* cannot be
    reached.  Small *timeout* values are suitable if *url* is an ip address.
    *kwargs* will be used to make :class:`urllib.request.Request` instance
    for opening the *url*."""

    try:
        from urllib2 import urlopen, URLError, Request
    except ImportError:
        from urllib.request import urlopen, Request
        from urllib.error import URLError

    if kwargs:
        request = Request(url, **kwargs)
    else:
        request = str(url)

    try:
        return urlopen(request, timeout=int(timeout))
    except URLError:
        raise IOError('{0} could not be opened for reading, invalid URL or '
                      'no internet connection'.format(repr(request)))


def glob(*pathnames):
    """Returns concatenation of ordered lists of paths matching patterns in
    *pathnames*."""

    paths = []
    for pathname in pathnames:
        matches = pyglob(pathname)
        matches.sort()
        paths.extend(matches)
    return paths


def copyFile(src, dst):
    """Returns *dst*, a copy of *src*."""

    try:
        copy(src, dst)
    except shError:
        return dst
    return dst


def addext(filename, extension):
    """Returns *filename*, with *extension* if it does not have one."""

    return filename + ('' if splitext(filename)[1] else extension)



from numpy import array, invert, ones, tile

from prody import Ensemble, PDBEnsemble, LOGGER
from prody.tests.datafiles import parseDatafile, DATA_FILES

LOGGER.verbosity = 'none'

ATOL = 1e-5
RTOL = 0

ATOMS = parseDatafile('multi_model_truncated', subset='ca')
ALLATOMS = parseDatafile('multi_model_truncated')
DCD = parseDatafile('dcd')
COORDS = ATOMS.getCoords()
COORDSETS = ATOMS.getCoordsets()
ENSEMBLE = Ensemble(ATOMS)
CONF = ENSEMBLE[0]
DATA = DATA_FILES['multi_model_truncated']
ENSEMBLE_RMSD = DATA['rmsd_ca']
ENSEMBLE_SUPERPOSE = DATA['rmsd_ca_aligned']

ENSEMBLEW = Ensemble(ATOMS)
ENSEMBLEW.setWeights(ones(len(ATOMS), dtype=float))
CONFW = ENSEMBLEW[0]

PDBENSEMBLE = PDBEnsemble('PDBEnsemble')
PDBENSEMBLE.setCoords(COORDS)
WEIGHTS = []
for i, xyz in enumerate(ATOMS.iterCoordsets()):
    weights = ones((len(ATOMS), 1), dtype=float)
    if i > 0:
        weights[i] = 0
        weights[-i] = 0
        PDBENSEMBLE.addCoordset(xyz, weights=weights)
    else:
        PDBENSEMBLE.addCoordset(xyz)
    WEIGHTS.append(weights)
PDBCONF = PDBENSEMBLE[0]
WEIGHTS = array(WEIGHTS)
WEIGHTS_BOOL = tile(WEIGHTS.astype(bool), (1,1,3))
WEIGHTS_BOOL_INVERSE = invert(WEIGHTS_BOOL)

PDBENSEMBLEA = PDBEnsemble('PDBEnsemble 2')
PDBENSEMBLEA.setAtoms(ATOMS)
PDBENSEMBLEA.setCoords(COORDS)
for i in range(ATOMS.numCoordsets()):
    weights = ones((len(ATOMS), 1), dtype=float)
    ATOMS.setACSIndex(i)
    if i > 0:
        weights[i] = 0
        weights[-i] = 0
        PDBENSEMBLEA.addCoordset(ATOMS, weights=weights, degeneracy=True)
    else:
        PDBENSEMBLEA.addCoordset(ATOMS, degeneracy=True)

ATOMS.setACSIndex(0)
